# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides classes for the (WS) SOAP I{document/literal}.
"""

from logging import getLogger
from suds import *
from suds.bindings.binding import Binding
from suds.sax.element import Element

log = getLogger(__name__)


class Document(Binding):
    """
    The document/literal style.  Literal is the only (@use) supported
    since document/encoded is pretty much dead.
    Although the soap specification supports multiple documents within the soap
    <body/>, it is very uncommon.  As such, suds presents an I{RPC} view of
    service methods defined with a single document parameter.  This is done so 
    that the user can pass individual parameters instead of one, single document.
    To support the complete specification, service methods defined with multiple documents
    (multiple message parts), must present a I{document} view for that method.
    """

    def __init__(self, wsdl):
        """
        @param wsdl: A WSDL object.
        @type wsdl: L{suds.wsdl.Definitions}
        """
        Binding.__init__(self, wsdl)
        
    def bodycontent(self, method, args, kwargs):
        """
        Get the content for the soap I{body} node.
        @param method: A service method.
        @type method: I{service.Method}
        @param args: method parameter values
        @type args: list
        @param kwargs: Named (keyword) args for the method invoked.
        @type kwargs: dict
        @return: The xml content for the <body/>
        @rtype: [L{Element},..]
        """
        n = 0
        pts = self.bodypart_types(method)
        if not len(pts):
            return ()
        root = self.document(pts)
        method.soap.input.body.root = root
        for pd in self.param_defs(method):
            if n < len(args):
                value = args[n]
            else:
                value = kwargs.get(pd[0])
            p = self.mkparam(method, pd, value)
            if p is not None:
                root.append(p)
            n += 1
        if len(pts) > 1:
            return root.children
        else:
            return root
        
    def document(self, pts):
        """
        Get the document root.  For I{document/literal}, this is the
        name of the wrapper element qualifed by the schema tns.
        @param pts: The method name.
        @type pts: str
        @return: A root element.
        @rtype: L{Element}
        """
        tag = pts[0][1].name
        ns = pts[0][1].namespace()
        d = Element(tag, ns=ns)
        return d
        
    def param_defs(self, method):
        """
        Get parameter definitions.
        @param method: A method name.
        @type method: basestring
        @return: A collection of parameter definitions
        @rtype: [(str, L{xsd.sxbase.SchemaObject}),..]
        """
        pts = self.bodypart_types(method)
        if len(pts) > 1:
            return pts
        result = []
        for p in pts:
            resolved = p[1].resolve()
            for child, ancestry in resolved:
                result.append((child.name, child))
        return result
    
    def returned_types(self, method):
        """
        Get the referenced type returned by the I{method}.
        @param method: The name of a method.
        @type method: str
        @return: The name of the type return by the method.
        @rtype: [L{xsd.sxbase.SchemaObject}]
        """
        result = []
        for pt in self.bodypart_types(method, input=False):
            pt = pt.resolve(nobuiltin=True)
            for child, ancestry in pt:
                result.append(child)
            break
        return result


#!/usr/bin/python
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

import sys
from setuptools import setup, find_packages
from suds import properties

setup(
    name="suds",
    version=properties['version'],
    description="Lightweight SOAP client",
    author="Jeff Ortel",
    author_email="jortel@redhat.com",
    maintainer="Jeff Ortel",
    maintainer_email="jortel@redhat.com",
    packages=find_packages(exclude=['tests']),
    url="https://fedorahosted.org/suds",
)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
The I{2nd generation} service proxy provides access to web services.
See I{README.txt}
"""

from logging import getLogger
import suds.metrics as metrics
from cookielib import CookieJar
from suds import *
from suds.transport import HttpAuthenticated, TransportError, Request
from suds.servicedefinition import ServiceDefinition
from suds import sudsobject
from sudsobject import Factory as InstFactory
from sudsobject import Object
from suds.resolver import PathResolver
from suds.builder import Builder
from suds.wsdl import Definitions
from suds.sax.document import Document
from suds.sax.parser import Parser
from suds.options import Options
from urlparse import urlparse

log = getLogger(__name__)


class Client(object):
    """ 
    A lightweight web services client.
    I{(2nd generation)} API.
    @ivar wsdl: The WSDL object.
    @type wsdl:L{Definitions}
    @ivar service: The service proxy used to invoke operations.
    @type service: L{Service}
    @ivar factory: The factory used to create objects.
    @type factory: L{Factory}
    @ivar sd: The service definition
    @type sd: L{ServiceDefinition}
    @ivar messages: The last sent/received messages.
    @type messages: str[2]
    """
    @classmethod
    def items(cls, sobject):
        """
        Extract the I{items} from a suds object much like the
        items() method works on I{dict}.
        @param sobject: A suds object
        @type sobject: L{Object}
        @return: A list of items contained in I{sobject}.
        @rtype: [(key, value),...]
        """
        return sudsobject.items(sobject)
    
    @classmethod
    def dict(cls, sobject):
        """
        Convert a sudsobject into a dictionary.
        @param sobject: A suds object
        @type sobject: L{Object}
        @return: A python dictionary containing the
            items contained in I{sobject}.
        @rtype: dict
        """
        return sudsobject.asdict(sobject)
    
    @classmethod
    def metadata(cls, sobject):
        """
        Extract the metadata from a suds object.
        @param sobject: A suds object
        @type sobject: L{Object}
        @return: The object's metadata
        @rtype: L{sudsobject.Metadata}
        """
        return sobject.__metadata__

    def __init__(self, url, **kwargs):
        """
        @param url: The URL for the WSDL.
        @type url: str
        @param kwargs: keyword arguments.
        @see: L{Options}
        """
        options = Options()
        options.transport = HttpAuthenticated(options)
        options.set(**kwargs)
        self.options = options
        self.wsdl = Definitions(url, options)
        self.service = Wrapper(Service(self))
        self.factory = Factory(self.wsdl)
        self.sd = ServiceDefinition(self.wsdl)
        self.messages = dict(tx=None, rx=None)
        
    def set_options(self, **kwargs):
        """
        Set options.
        @param kwargs: keyword arguments.
        @see: L{Options}
        """
        self.options.set(**kwargs)
        
    def add_prefix(self, prefix, uri):
        """
        Add I{static} mapping of an XML namespace prefix to a namespace.
        This is useful for cases when a wsdl and referenced schemas make heavy
        use of namespaces and those namespaces are subject to changed.
        @param prefix: An XML namespace prefix.
        @type prefix: str
        @param uri: An XML namespace URI.
        @type uri: str
        @raise Exception: when prefix is already mapped.
        """
        root = self.wsdl.root
        mapped = root.resolvePrefix(prefix, None)
        if mapped is None:
            root.addPrefix(prefix, uri)
            return
        if mapped[1] != uri:
            raise Exception('"%s" already mapped as "%s"' % (prefix, mapped))
        
    def last_sent(self):
        """
        Get last sent I{soap} message.
        @return: The last sent I{soap} message.
        @rtype: L{Document}
        """
        return self.messages.get('tx')
    
    def last_received(self):
        """
        Get last received I{soap} message.
        @return: The last received I{soap} message.
        @rtype: L{Document}
        """
        return self.messages.get('rx')
 
    def __str__(self):
        return unicode(self)
        
    def __unicode__(self):
        s = ['\n']
        version = properties.get('version')
        build = properties.get('build').split()
        s.append('Suds ( https://fedorahosted.org/suds/ )')
        s.append('  version: %s' % version)
        s.append(' %s  build: %s' % (build[0], build[1]))
        s.append('\n\n%s' % unicode(self.sd))
        return ''.join(s)


class Factory:
    """
    A factory for instantiating types defined in the wsdl
    @ivar resolver: A schema type resolver.
    @type resolver: L{PathResolver}
    @ivar builder: A schema object builder.
    @type builder: L{Builder}
    """
    
    def __init__(self, wsdl):
        """
        @param wsdl: A schema object.
        @type wsdl: L{wsdl.Definitions}
        """
        self.resolver = PathResolver(wsdl)
        self.builder = Builder(self.resolver)
    
    def create(self, name):
        """
        create a WSDL type by name
        @param name: The name of a type defined in the WSDL.
        @type name: str
        @return: The requested object.
        @rtype: L{Object}
        """
        timer = metrics.Timer()
        timer.start()
        type = self.resolver.find(name)
        if type is None:
            raise TypeNotFound(name)
        if type.enum():
            result = InstFactory.object(name)
            for e, a in type.children():
                setattr(result, e.name, e.name)
        else:
            try:
                result = self.builder.build(type)
            except Exception, e:
                log.error("create '%s' failed", name, exc_info=True)
                raise BuildError(name, e)
        timer.stop()
        metrics.log.debug('%s created: %s', name, timer)
        return result

   
class Wrapper:
    """
    Wrapper, translates to resolve() and call().
    """
    
    def __init__(self, wrapped):
        self.__wrapped__ = wrapped

    def __getattr__(self, name):
        builtin =  name.startswith('__') and name.endswith('__')
        if builtin:
            return self.__dict__[name]
        else:
            return self.__wrapped__.resolve(name)
        
    def __call__(self, *args, **kwargs):
        target = self.__wrapped__
        return target.call(*args, **kwargs)
    
    def __str__(self):
        return str(self.__wrapped__)
    
    def __unicode__(self):
        return unicode(self.__wrapped__) 
        

class Service:
    """
    The I{service} (namespace) object.
    @ivar client: A client object.
    @type client: L{Client}
    """
    
    def __init__(self, client):
        """
        @param client: A client object.
        @type client: L{Client}
        """
        self.client = client
        
    def resolve(self, name):
        """
        Resolve I{name} to a service port.
        @param name: A port/method name.
        @type name: str
        @return: The I{wrapped} L{Port}
        """
        service = self.client.wsdl.service
        port = self.dport(service)
        if port is None:
            port = service.port(name)
        return Wrapper(Port(self.client, (name, port)))
    
    def dport(self, service):
        """
        The I{default} port as defined by L{Options}.
        @param service: A wsdl service object.
        @type service : L{wsdl.Service}
        @return: The I{raw} L{wsdl.Port}
        """
        name = self.client.options.port
        if name is not None:
            port = service.port(name)
            if port is None:
                raise PortNotFound(name)
        else:
            port = None
        return port
    
    def __str__(self):
        return unicode(self)
    
    def __unicode__(self):
        return unicode(self.client.sd) 

        
class Port:
    """
    The I{port} (namespace) object.
    @ivar client: A client object.
    @type client: L{Client}
    @ivar port: A port tuple (name, L{wsdl.Port})
    @type port: tuple
    """
    
    def __init__(self, client, port):
        """
        @param client: A client object.
        @type client: L{Client}
        @param port: A port tuple (name, L{wsdl.Port})
        @type port: tuple
        """
        self.client = client
        self.port = port
        
    def resolve(self, name, strict=True):
        """
        Resolve I{name} to a service method.
        @param name: A method name.
        @type name: str
        @param strict: Make sure this is a real port.
        @type strict: bool
        @return: The I{wrapped} L{Method}
        """
        if strict and self.anyport():
            raise PortNotFound(self.name())
        finder = self.finder()
        method = finder.method(name)
        if method is None:
            raise MethodNotFound(name)
        return Wrapper(Method(self.client, method))
    
    def name(self):
        """
        The port name.
        @return: The port name.
        @rtype: str
        """
        return self.port[0]
    
    def realport(self):
        """
        The I{real} port.
        @return: The contained port object.
        @rtype: L{wsdl.Port}
        """
        return self.port[1]
    
    def anyport(self):
        """
        Get whether this port references a real port or just a
        placeholder for I{any} port.
        @return: True if real port reference is None.
        @rtype: bool
        """
        return ( self.realport() is None )
    
    def finder(self):
        """
        The method name I{finder}.
        @return: A method name resolver.
        @rtype: (L{wsdl.Port}|L{wsdl.Service})
        """
        if self.anyport():
            return self.client.wsdl.service
        else:
            return self.realport()

    def call(self, *args, **kwargs):
        """
        When called, this means that this port was returned by the
        Service.resolve() but the user intended to call a method.
        The name is used to lookup a method and forwards the invocation
        to the method.
        """
        method = self.resolve(self.name(), strict=False)
        return method(*args, **kwargs)


class Method:
    """
    The I{method} (namespace) object.
    @ivar client: A client object.
    @type client: L{Client}
    @ivar method: A I{wsdl} method.
    @type I{wsdl} Method.
    """

    def __init__(self, client, method):
        """
        @param client: A client object.
        @type client: L{Client}
        @param method: A I{raw} method.
        @type I{raw} Method.
        """
        self.client = client
        self.method = method
        
    def resolve(self, name):
        """
        Error, not permitted.
        @param name: A method attribute to be resolved.
        @type name: str.
        @raise AttributeError: Always.
        """
        raise AttributeError(name)
    
    def call(self, *args, **kwargs):
        """
        Method Invocation.
        """
        clientclass = self.clientclass(kwargs)
        client = clientclass(self.client, self.method)
        if not self.faults():
            try:
                return client.invoke(args, kwargs)
            except WebFault, e:
                return (500, e)
        else:
            return client.invoke(args, kwargs)
        
    def faults(self):
        return self.client.options.faults
        
    def clientclass(self, kwargs):
        if SimClient.simulation(kwargs):
            return SimClient
        else:
            return SoapClient


class SoapClient:
    """
    A lightweight soap based web client B{**not intended for external use}
    @ivar service: The target method.
    @type service: L{Service}
    @ivar method: A target method.
    @type method: L{Method}
    @ivar options: A dictonary of options.
    @type options: dict
    @ivar cookiejar: A cookie jar.
    @type cookiejar: libcookie.CookieJar
    """

    def __init__(self, client, method):
        """
        @param client: A suds client.
        @type client: L{Client}
        @param method: A target method.
        @type method: L{Method}
        """
        self.client = client
        self.method = method
        self.options = client.options
        self.cookiejar = CookieJar()
        
    def invoke(self, args, kwargs):
        """
        Send the required soap message to invoke the specified method
        @param args: A list of args for the method invoked.
        @type args: list
        @param kwargs: Named (keyword) args for the method invoked.
        @type kwargs: dict
        @return: The result of the method invocation.
        @rtype: I{builtin}|I{subclass of} L{Object}
        """
        timer = metrics.Timer()
        timer.start()
        result = None
        binding = self.method.binding.input
        binding.options = self.options
        msg = binding.get_message(self.method, args, kwargs)
        timer.stop()
        metrics.log.debug(
                "message for '%s' created: %s",
                self.method.qname, timer)
        timer.start()
        result = self.send(msg)
        timer.stop()
        metrics.log.debug(
                "method '%s' invoked: %s",
                self.method.qname, timer)
        return result
    
    def send(self, msg):
        """
        Send soap message.
        @param msg: A soap message to send.
        @type msg: basestring
        @return: The reply to the sent message.
        @rtype: I{builtin} or I{subclass of} L{Object}
        """
        result = None
        location = self.location()
        binding = self.method.binding.input
        transport = self.options.transport
        log.debug('sending to (%s)\nmessage:\n%s', location, msg)
        try:
            self.last_sent(Document(msg))
            request = Request(location, msg)
            request.headers = self.headers()
            reply = transport.send(request)
            result = self.succeeded(binding, reply)
        except TransportError, e:
            if e.httpcode in (202,204):
                result = None
            else:
                log.error(self.last_sent())
                result = self.failed(binding, e)
        return result
    
    def headers(self):
        """
        Get http headers or the http/https request.
        @return: A dictionary of header/values.
        @rtype: dict
        """
        action = self.method.soap.action
        stock = { 'Content-Type' : 'text/xml', 'SOAPAction': action }
        result = dict(stock, **self.options.headers)
        log.debug('headers = %s', result)
        return result
    
    def succeeded(self, binding, reply):
        """
        Request succeeded, process the reply
        @param binding: The binding to be used to process the reply.
        @type binding: L{bindings.binding.Binding}
        @return: The method result.
        @rtype: I{builtin}, L{Object}
        @raise WebFault: On server.
        """
        log.debug('http succeeded:\n%s', reply)
        if len(reply) > 0:
            r, p = binding.get_reply(self.method, reply)
            self.last_received(r)
            if self.options.faults:
                return p
            else:
                return (200, p)
        else:
            if self.options.faults:
                return None
            else:
                return (200, None)
        
    def failed(self, binding, error):
        """
        Request failed, process reply based on reason
        @param binding: The binding to be used to process the reply.
        @type binding: L{suds.bindings.binding.Binding}
        @param error: The http error message
        @type error: L{transport.TransportError}
        """
        status, reason = (error.httpcode, tostr(error))
        reply = error.fp.read()
        log.debug('http failed:\n%s', reply)
        if status == 500:
            if len(reply) > 0:
                r, p = binding.get_fault(reply)
                self.last_received(r)
                return (status, p)
            else:
                return (status, None)
        if self.options.faults:
            raise Exception((status, reason))
        else:
            return (status, None)

    def location(self):
        return self.options.get('location', self.method.location)
    
    def last_sent(self, d=None):
        key = 'tx'
        messages = self.client.messages
        if d is None:
            return messages.get(key)
        else:
            messages[key] = d
        
    def last_received(self, d=None):
        key = 'rx'
        messages = self.client.messages
        if d is None:
            return messages.get(key)
        else:
            messages[key] = d


class SimClient(SoapClient):
    """
    Loopback client used for message/reply simulation.
    """
    
    injkey = '__inject'
    
    @classmethod
    def simulation(cls, kwargs):
        """ get whether loopback has been specified in the I{kwargs}. """
        return kwargs.has_key(SimClient.injkey)
        
    def invoke(self, args, kwargs):
        """
        Send the required soap message to invoke the specified method
        @param args: A list of args for the method invoked.
        @type args: list
        @param kwargs: Named (keyword) args for the method invoked.
        @type kwargs: dict
        @return: The result of the method invocation.
        @rtype: I{builtin} or I{subclass of} L{Object}
        """
        simulation = kwargs[self.injkey]
        msg = simulation.get('msg')
        reply = simulation.get('reply')
        fault = simulation.get('fault')
        if msg is None:
            if reply is not None:
                return self.__reply(reply)
            if fault is not None:
                return self.__fault(fault)
            raise Exception('(reply|fault) expected when msg=None')
        msg = Parser().parse(string=msg)
        return self.send(msg)
    
    def __reply(self, reply):
        """ simulate the reply """
        binding = self.method.binding.output
        binding.options = self.options
        return self.succeeded(binding, reply)
    
    def __fault(self, reply):
        """ simulate the (fault) reply """
        binding = self.method.binding.output
        binding.options = self.options
        if self.options.faults:
            r, p = binding.get_fault(reply)
            self.last_received(r)
            return (500, p)
        else:
            return (500, None)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

def assertEquals(one, two):
    if one != two:
        raise Exception('Assertion Exception!  %s does not equal %s' % (one, two))

from suds.xsd.sxbuiltin import *
import unittest
import time
    
class DateTest(unittest.TestCase):
    def setUp(self):
        self.tz = time.timezone/60/60
        
    def getTestersTimezoneString(self):

        offset = str(self.tz)
        if int(offset) >= 0:
            if len(offset) == 2:
                offset = "+%s:00" % offset
            else:
                offset = "+0%s:00" % offset
        else:
            if len(offset) == 2:
                offset = "-%s:00" % offset
            else:
                offset = "-0%s:00" % offset
                
        return offset
                
    def getDay(self, day, hour, current_tz):
        if (current_tz - self.tz + hour) > 24:
            return day + 1
        elif (current_tz - self.tz + hour) < 0:
            return day - 1
        else:
            return day
            
    def getHour(self, hour, current_tz):
        new_hour = (current_tz - self.tz + hour)
        if new_hour > 23:
            return new_hour - 24
        elif new_hour < 0:
            return 24 - (new_hour*-1)
        else:
            return new_hour
            
    def getHourString(self, hour, current_tz):
        hour = self.getHour(hour, current_tz)
        
        if len(str(hour)) > 1:
            return "%s" % hour
        else:
            return "0%s" % hour
            
class TestXDate(DateTest):
    """
    examples
    2002-09-24
    2002-09-24Z
    2002-09-24-06:00
    2002-09-24+06:00
    """
    #fake it out a bit so I can just test what matters
    def fake_init(self):
        pass
    XDate.__init__ = fake_init
    
    def test_should_return_correct_date_object_given_simple_date(self):
        date = XDate().translate("2006-10-10")
        self.assertEqual(date.day, 10)
        self.assertEqual(date.month, 10)
        self.assertEqual(date.year, 2006)
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.hour, 0)
        
    def test_should_return_correct_string_from_date_object_given_simple_date(self):
        date = XDate().translate(XDate().translate("2006-10-10"), False)
        
        self.assertEquals("2006-10-10", date)
        
    def test_should_return_correct_date_object_given_date_with_timezone(self):
        date = XDate().translate("1945-08-20+06:00")
        self.assertEqual(date.day, self.getDay(20, 0, 6))
        self.assertEqual(date.month, 8)
        self.assertEqual(date.year, 1945)
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.hour, self.getHour(0, 6))
        
    def test_should_return_correct_string_from_date_object_given_date_with_timezone(self):
        date = XDate().translate(XDate().translate("1945-08-20+06:00"), False)
        
        self.assertEquals("1945-08-%s%s" % (str(self.getDay(20, 0, 6)), self.getTestersTimezoneString()), date)

    def test_should_return_correct_date_object_given_date_with_different_timezone(self):
        date = XDate().translate("1945-08-20+03:00")
        self.assertEqual(date.day, self.getDay(20, 0, 3))
        self.assertEqual(date.month, 8)
        self.assertEqual(date.year, 1945)
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.hour, self.getHour(0, 3))

    def test_should_return_correct_string_from_date_object_given_date_with_timezone(self):
        date = XDate().translate(XDate().translate("1945-08-20+03:00"), False)

        self.assertEquals("1945-08-%s%s" % (str(self.getDay(20, 0, 3)), self.getTestersTimezoneString()), date)

    def test_should_return_correct_date_object_given_date_with_different_utc(self):
        #from where I am, this is 6 hours off and would become a different day
        date = XDate().translate("1945-08-20Z")
        self.assertEqual(date.day, self.getDay(20, 0, 0))
        self.assertEqual(date.month, 8)
        self.assertEqual(date.year, 1945)
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.hour, self.getHour(0, 0))

    def test_should_return_correct_string_from_date_object_given_date_with_utc(self):
        date = XDate().translate(XDate().translate("1945-08-20Z"), False)

        self.assertEquals("1945-08-%s%s" % (str(self.getDay(20, 0, 0)), self.getTestersTimezoneString()), date)
        
    def test_null_does_not_throw_exception(self):
        self.assertEquals("", XDate().translate(None, False))
      
class TestXTime(DateTest):
    """
    09:00:00
    09:30:10.5
    09:30:10Z
    09:30:10-06:00
    09:30:10+06:00
    """
    #fake it out a bit so I can just test what matters
    def fake_init(self):
        pass
    XTime.__init__ = fake_init

    def test_should_return_correct_time_object_given_simple_time(self):
        date = XTime().translate("09:00:00")
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.hour, 9)

    def test_should_return_correct_string_from_time_object_given_simple_time(self):
        date = XTime().translate(XTime().translate("09:00:00"), False)

        self.assertEquals("09:00:00", date)
        
    def test_should_return_correct_time_object_given_time_with_microseconds(self):
        time = XTime().translate("09:30:10.5")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, 9)
        self.assertEqual(time.microsecond, 500000)

    def test_should_return_correct_string_from_time_object_given_time_with_microseconds(self):
        date = XTime().translate(XTime().translate("09:30:10.5"), False)

        self.assertEquals("09:30:10.5", date)

    def test_should_return_correct_time_object_given_time_with_utc(self):
        time = XTime().translate("09:30:10Z")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(9, 0))
        self.assertEqual(time.microsecond, 0)

    def test_should_return_correct_string_from_time_object_given_time_with_utc(self):
        date = XTime().translate(XTime().translate("09:30:10Z"), False)

        self.assertEquals("%s:30:10%s" % (self.getHourString(9, 0), self.getTestersTimezoneString()), date)
                
    def test_should_return_correct_time_object_given_complex_time(self):
        time = XTime().translate("09:30:10.525+09:00")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(9, 9))
        self.assertEqual(time.microsecond, 525000)

    def test_should_return_correct_string_from_time_object_given_complex_time(self):
        date = XTime().translate(XTime().translate("09:30:10.525+09:00"), False)

        self.assertEquals("%s:30:10.525%s" % (self.getHourString(9, 9), self.getTestersTimezoneString()), date)
        
    def test_should_wrap_time_correctly(self):
        time = XTime().translate("23:30:10.525+09:00")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(23, 9))
        self.assertEqual(time.microsecond, 525000)
        
        time = XTime().translate("00:30:10.525-03:00")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(0, -3))
        self.assertEqual(time.microsecond, 525000)
        
        time = XTime().translate("05:30:10.525+03:00")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(5, 3))
        self.assertEqual(time.microsecond, 525000)
        
        time = XTime().translate("05:30:10.525-03:00")
        self.assertEqual(time.minute, 30)
        self.assertEqual(time.second, 10)
        self.assertEqual(time.hour, self.getHour(5, -3))
        self.assertEqual(time.microsecond, 525000)

    def test_null_does_not_throw_exception(self):
        self.assertEquals("", XTime().translate(None, False))
        
class TestXDateTime(DateTest):
    """
    2002-05-30T09:00:00
    2002-05-30T09:30:10.5
    """
    #fake it out a bit so I can just test what matters
    def fake_init(self):
        pass
    XDateTime.__init__ = fake_init

    def test_should_return_correct_time_object_given_simple_time(self):
        date = XDateTime().translate("2002-05-30T09:00:00")
        self.assertEqual(date.minute, 0)
        self.assertEqual(date.second, 0)
        self.assertEqual(date.microsecond, 0)
        self.assertEqual(date.hour, 9)
        self.assertEqual(date.day, 30)
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)

    def test_should_return_correct_string_from_time_object_given_simple_time(self):
        date = XDateTime().translate(XDateTime().translate("2002-05-30T09:00:00"), False)

        self.assertEquals("2002-05-30T09:00:00", date)
        
    def test_should_return_correct_time_object_given_simple_time(self):
        date = XDateTime().translate("2002-05-30T09:30:10.5")
        self.assertEqual(date.minute, 30)
        self.assertEqual(date.second, 10)
        self.assertEqual(date.microsecond, 500000)
        self.assertEqual(date.hour, 9)
        self.assertEqual(date.day, 30)
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)

    def test_should_return_correct_string_from_time_object_given_simple_time(self):
        date = XDateTime().translate(XDateTime().translate("2002-05-30T09:30:10.5"), False)

        self.assertEquals("2002-05-30T09:30:10.5", date)
        
    def test_should_wrap_time_correctly(self):
        date = XDateTime().translate("2002-05-30T23:30:10.525+09:00")
        self.assertEqual(date.minute, 30)
        self.assertEqual(date.second, 10)
        self.assertEqual(date.hour, self.getHour(23, 9))
        self.assertEqual(date.microsecond, 525000)
        self.assertEqual(date.day, self.getDay(30, 23, 9))
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)

        date = XDateTime().translate("2002-05-30T00:30:10.525-03:00")
        self.assertEqual(date.minute, 30)
        self.assertEqual(date.second, 10)
        self.assertEqual(date.hour, self.getHour(0, -3))
        self.assertEqual(date.microsecond, 525000)
        self.assertEqual(date.day, self.getDay(30, 0, -3))
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)

        date = XDateTime().translate("2002-05-30T05:30:10.525+03:00")
        self.assertEqual(date.minute, 30)
        self.assertEqual(date.second, 10)
        self.assertEqual(date.hour, self.getHour(5, 3))
        self.assertEqual(date.microsecond, 525000)
        self.assertEqual(date.day, self.getDay(30, 5, 3))
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)

        date = XDateTime().translate("2002-05-30T05:30:10.525-03:00")
        self.assertEqual(date.minute, 30)
        self.assertEqual(date.second, 10)
        self.assertEqual(date.hour, self.getHour(5, -3))
        self.assertEqual(date.microsecond, 525000)
        self.assertEqual(date.day, self.getDay(30, 5, -3))
        self.assertEqual(date.year, 2002)
        self.assertEqual(date.month, 5)
        
    def test_null_does_not_throw_exception(self):
        self.assertEquals("", XDateTime().translate(None, False))
        
if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the 
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Contains transport interface (classes) and reference implementation.
"""

from logging import getLogger

log = getLogger(__name__)


#
# Transport Interface
#


class TransportError(Exception):
    def __init__(self, reason, httpcode, fp=None):
        Exception.__init__(self, reason)
        self.httpcode = httpcode
        self.fp = fp

class Request:
    """
    A transport request
    @ivar url: The url for the request.
    @type url: str
    @ivar proxy: The url of the proxy to be used for the request.
    @type proxy: str
    @ivar message: The message to be sent in a POST request.
    @type message: str
    @ivar headers: The http headers to be used for the request.
    @type headers: dict
    """

    def __init__(self, url, message=None):
        """
        @param url: The url for the request.
        @type url: str
        @param message: The (optional) message to be send in the request.
        @type message: str
        """
        self.url = url
        self.headers = {}
        self.message = message

class Transport:
    """
    The transport I{interface}.
    """
    
    def __init__(self, options=None):
        """
        @param options: A suds options object.
        @type options: L{suds.options.Options}
        """
        if options is None:
            from suds.options import Options
            self.options = Options()
            del Options
        else:
            self.options = options
    
    def open(self, request):
        """
        Open the url in the specified request.
        @param request: A transport request.
        @type request: L{Request}
        @return: An input stream.
        @rtype: stream
        @raise TransportError: On all transport errors.
        """
        raise Exception('not-implemented')
    
    def send(self, request):
        """
        Send soap message.  Implementations are expected to handle:
            - proxies
            - I{http} headers
            - cookies
            - sending message
            - brokering exceptions into L{TransportError}
        @param request: A transport request.
        @type request: L{Request}
        @return: The reply
        @rtype: basestring
        @raise TransportError: On all transport errors.
        """
        raise Exception('not-implemented')


#
# Transport Implementation
#

import urllib2 as u2
from urlparse import urlparse
from cookielib import CookieJar

class HttpTransport(Transport):
    """
    urllib2 transport implementation.
    """
    
    def __init__(self, options=None):
        Transport.__init__(self, options)
        self.cookiejar = CookieJar()
        self.urlopener = None
        
    def open(self, request):
        try:
            url = request.url
            log.debug('opening (%s)', url)
            u2request = u2.Request(url)
            self.__setproxy(url, u2request)
            return self.__open(u2request)
        except u2.HTTPError, e:
            raise TransportError(str(e), e.code, e.fp)

    def send(self, request):
        result = None
        url = request.url
        msg = str(request.message)
        headers = request.headers
        log.debug('sending to (%s)\nmessage:\n%s', url, msg)
        try:
            u2request = u2.Request(url, msg, headers)
            self.__addcookies(u2request)
            self.__setproxy(url, u2request)
            fp = self.__open(u2request)
            self.__getcookies(fp, u2request)
            result = fp.read()
        except u2.HTTPError, e:
            if e.code in (202,204):
                result = None
            else:
                raise TransportError(e.msg, e.code, e.fp)
        return result

    def __addcookies(self, u2request):
        self.cookiejar.add_cookie_header(u2request)
        
    def __getcookies(self, fp, u2request):
        self.cookiejar.extract_cookies(fp, u2request)
        
    def __open(self, u2request):
        if self.urlopener is None:
            return u2.urlopen(u2request)
        else:
            return self.urlopener.open(u2request)
        
    def __setproxy(self, url, u2request):
        protocol = urlparse(url)[0]
        proxy = self.options.proxy.get(protocol, None)
        if proxy is None:
            return
        protocol = u2request.type
        u2request.set_proxy(proxy, protocol)
        

class HttpAuthenticated(HttpTransport):
    """
    Provides basic http authentication.
    @ivar pm: The password manager.
    @ivar handler: The authentication handler.
    """
    
    def __init__(self, options):
        HttpTransport.__init__(self, options)
        self.pm = u2.HTTPPasswordMgrWithDefaultRealm()
        self.handler = u2.HTTPBasicAuthHandler(self.pm)
        self.urlopener = u2.build_opener(self.handler)
        
    def open(self, request):
        self.__addcredentials(request)
        return  HttpTransport.open(self, request)

    def send(self, request):
        self.__addcredentials(request)
        return HttpTransport.send(self, request)
    
    def __addcredentials(self, request):
        user = self.options.username
        pwd = self.options.password
        if user is not None:
            self.pm.add_password(None, request.url, user, pwd)
            
